@@ -5,27 +5,26 @@
 
 from concurrent.futures import as_completed, ThreadPoolExecutor
 from glob import glob
+import json
 import os
-
 from tqdm import tqdm
 
 
 class Source():
+    """
+    A data structure representing information about a subset of a collection.
+    V1 collections have one source, V2+ can contain multiple each with their
+    own imagesets.
+    """
+
+    def __repr__(self):
+        return '<Source "{}" from Collection "{}", id: "{}"'.format(
+            self.name, self.collection.name, self.id)
 
     def __init__(self, collection, source_dict):
         self._collection = collection
         self._data = source_dict
 
-    @property
-    def name():
-        pass
-
-    @name.getter
-    def name(self):
-        assert self._data, 'Source had no self._data set'
-        assert 'name' in self._data.keys(), 'Source\'s data didn\'t have a \'name\' key'
-        return self._data['name']
-
     @property
     def collection():
         pass
@@ -34,15 +33,25 @@ class Source():
     def collection(self):
         return self._collection
 
+    @property
+    def name():
+        pass
+
+    @name.getter
+    def name(self):
+        return self._retrieve('name')
+
     @property
     def id():
         pass
 
     @id.getter
     def id(self):
-        assert self._data, 'Source had no self._data set'
-        assert 'source_id' in self._data, 'Source\'s data didn\'t have a \'source_id\' key'
-        return self._data['source_id']
+        """The .source_id of this Source. Note: Invalid for V1 collections."""
+
+        if self.collection.version < 2:
+            return None
+        return self._retrieve('source_id')
 
     @property
     def imageset_id():
@@ -50,9 +59,19 @@ class Source():
 
     @imageset_id.getter
     def imageset_id(self):
-        assert self._data, 'Source had no self._data set'
-        assert 'imageset_id' in self._data, 'Source\'s data didn\'t have an \'imageset_id\' key'
-        return self._data['imageset_id']
+        return self._retrieve('imageset_id')
+
+    @property
+    def index():
+        pass
+
+    @index.getter
+    def index(self) -> int:
+        """
+        The index/position of this source in its collection's .sources list.
+        """
+
+        return self.collection.sources.index(self)
 
     @property
     def _imageset_dataset_join_id():
@@ -60,10 +79,28 @@ class Source():
 
     @_imageset_dataset_join_id.getter
     def _imageset_dataset_join_id(self):
-        k = 'imageset_dataset_join_id'
-        assert self._data, 'Source had no self._data set'
-        assert k in self._data, 'Source\'s data didn\'t have an \'{}\' key'.format(k)
-        return self._data[k]
+        return self._retrieve('imageset_dataset_join_id')
+
+    def _retrieve(self, key):
+        if key not in self._data:
+            raise KeyError('Key "{}" not found in Source _data'.format(key))
+        return self._data[key]
+
+    @property
+    def image_details():
+        pass
+
+    @image_details.getter
+    def image_details(self):
+
+        collection = self.collection
+        c = collection.client
+
+        ims_url = '{}/{}/project/{}/nodes/{}/images'.format(
+            c.HOME, c.API_1, collection.workspace_id, self.imageset_id)
+        ims = c._auth_get(ims_url)
+
+        return ims
 
 
 class UploadableSource():
@@ -93,12 +130,24 @@ class UploadableSource():
         ".json"
     )
 
-    def __init__(self, name, image_dir, column_filename='Filename', recursive_search=True):
-        """Used in conjunction with create_collection().
-
-        An UploadableSource() points towards and manages the upload of local files, resulting in the
-        generation of a true Source() in the collection.
+    def __init__(self, name, image_dir, column_filename='__auto_join__', recursive_search=True, filename_filter=[],
+                 additional_mimes={}):
         """
+        Used in conjunction with create_collection().
+
+        An UploadableSource() points towards and manages the upload of local
+        files, resulting in the generation of a true Source() in the
+        collection.
+
+        To limit to an allowed specific list of filenames, provide
+        'filename_filter'. This filter will check against
+        os.path.basename(filepath).
+
+        Common mime types are inferred from the file extension,
+        but a dict of additional mime type mappings can be provided eg to
+        cater for files with no extension.
+        """
+
         self.name = name
         self.image_dir = image_dir
         self.column_filename = column_filename
@@ -107,17 +156,27 @@ class UploadableSource():
         self._source = None
         self._index = None
 
+        self.image_mimes = {**UploadableSource.IMAGE_MIMES, **additional_mimes}
+
         # Check the directory exists
         if not os.path.exists(image_dir):
             raise FileNotFoundError('image_dir "{}" does not exist'.format(self.image_dir))
         if not os.path.isdir(image_dir):
             raise TypeError('image_dir "{}" is not a directory'.format(self.image_dir))
 
-        # Find all files matching the allowed mime-types
-        self.filepaths = sum(
-            [glob('{}/**/*{}'.format(image_dir, ext), recursive=recursive_search)
-                for ext in self.IMAGE_MIMES.keys()], [])
+        # Potentially limit paths based on filename_filter.
+        if filename_filter:
+            if type(filename_filter) != list:
+                raise TypeError('filename_filter should be a list')
+            fps = [os.path.join(image_dir, fp) for fp in filename_filter if os.path.exists(os.path.join(image_dir, fp))]
 
+        else:
+            # Find all files matching the allowed mime-types
+            fps = sum(
+                [glob('{}/**/*{}'.format(image_dir, ext), recursive=recursive_search)
+                 for ext in self.IMAGE_MIMES.keys()], [])
+
+        self.filepaths = fps
         self.filenames = [os.path.basename(fp) for fp in self.filepaths]
 
         print('UploadableSource "{}" found {} images in "{}"'.format(self.name, len(self), image_dir))
@@ -177,19 +236,19 @@ class UploadableSource():
         self._index = index
         self._source = source
 
-        if not self.source.name == self.name:
+        if not self.source.name == 'None' and not self.source.name == self.name:
             raise Exception(
                 'UploadableSource "{}" registered to Source "{}" when their names should match'
                 .format(self.name, self.source.name)
             )
 
-    def _assign_images_to_smaller_lists(self, file_paths):
+    def _assign_images_to_smaller_lists(self, file_paths, start=0):
         """Create smaller lists based on the number of images in the directory."""
         # Recurse and pick up only valid files (either with image extensions, or not on blacklist)
         total_work = len(file_paths)
         workloads = []
         workload = []
-        start = 0
+        workload_start = start
 
         if total_work > 2500:
             size = 100
@@ -204,9 +263,9 @@ class UploadableSource():
             workload.append(path)
             i += 1
             if len(workload) == size or i == total_work:
-                workloads.append({'paths': workload, 'start': start})
+                workloads.append({'paths': workload, 'start': workload_start})
                 workload = []
-                start = i
+                workload_start = start + i
 
         return workloads, total_work, size
 
@@ -232,9 +291,16 @@ class UploadableSource():
 
         # Tell the server how many uploads are expected for this source
         url = '{}/{}/project/{}/imagesets/{}/extend'.format(c.HOME, c.API_0, collection.workspace_id, self.imageset_id)
-        c._auth_post(url, body=None, json={'delta': len(self)})
+        delta = len(self)
+        # If there are no new uploads, ignore.
+        if delta == 0:
+            print('No new data to be uploaded.')
+            return
+        resp = c._auth_post(url, body=None, json={'delta': delta})
+        new_size = resp['new_size']
+        start = new_size - delta
 
-        (workloads, total_work, group_size) = self._assign_images_to_smaller_lists(self.filepaths)
+        (workloads, total_work, group_size) = self._assign_images_to_smaller_lists(self.filepaths, start=start)
 
         # Multiprocess upload the images
         # divide the filepaths into smaller groups
@@ -265,7 +331,7 @@ class UploadableSource():
 
         # Obtain blob storage information
         blob_storage_urls, id_set = c._obtain_signed_blob_storage_urls(
-            coll.workspace_id, id_count=len(paths))
+            coll.workspace_id, id_count=len(paths), blob_path="imagesets/{}".format(self.imageset_id))
 
         # Check that numbers of values are still matching
         if not len(paths) == len(blob_storage_urls):
@@ -304,7 +370,7 @@ class UploadableSource():
 
     def _check_in_data(self, data):
         cols = list(data.columns)
-        if self.column_filename not in cols:
+        if self.column_filename != '__auto_join__' and self.column_filename not in cols:
             raise Exception('Source "{}" had the filename_column "{}" '
                             'which is not a column of the provided data:\n{}'
                             .format(self.name, self.column_filename, cols))
@@ -328,10 +394,60 @@ class UploadableSource():
 
         return uploadable_sources
 
-    @classmethod
-    def _get_mime_type(cls, path) -> str:
+    def _get_mime_type(self, path) -> str:
         """Gets the mime_type of the path. Raises an error if not a valid image mime_type."""
+        if '.' not in path:
+            return self.image_mimes['']
         ext = os.path.splitext(path)[-1]
-        if ext in cls.IMAGE_MIMES.keys():
-            return cls.IMAGE_MIMES[ext]
-        raise TypeError('"{}" is not a supported image mime_type ({})'.format(path, cls.IMAGE_MIMES))
+        if ext in self.image_mimes.keys():
+            return self.image_mimes[ext]
+        raise TypeError('"{}" is not a supported image mime_type ({})'.format(path, self.image_mimes))
+
+
+class UrlSource(UploadableSource):
+
+    def __init__(self, name, url_template, image_fetch_headers, column_filename=None):
+        """Used in conjunction with create_collection().
+
+        A UrlSource() fetches the images from the url template given, resulting in the
+        generation of a true Source() in the collection.
+        """
+        self.name = name
+        self.url_template = url_template
+        self.image_fetch_headers = image_fetch_headers
+        self.column_filename = column_filename
+
+        # Set externally once a blank collection has been made
+        self._source = None
+        self._index = None
+
+    def _upload(self):
+        """Update upload imageset to use the provided url template to get the images.
+
+        provided a Source() has been generated and designated to this instance.
+        """
+        collection = self.source.collection
+        c = collection.client
+
+        print('- Configuring source {} "{}" to fetch images from url'
+              .format(self.index, self.name))
+
+        upload_ims_url = '{}/{}/project/{}/imagesets/{}'.format(
+            c.HOME, c.API_0, collection.workspace_id, self.imageset_id)
+        upload_ims = c._auth_get(upload_ims_url)
+
+        new_source = {
+            "dataset_id": collection._dataset_id,
+            'fetch': {
+                'headers': self.image_fetch_headers,
+                'url': {
+                    'dataset_column': self.column_filename,
+                    'url_template': self.url_template,
+                }
+            }
+        }
+        upload_ims['imageset']['source'] = new_source
+        payload = json.dumps(upload_ims['imageset'])
+        r = c._auth_put(upload_ims_url, payload, return_response=True)
+
+        return r
